/*

Copyright (c) 2002                      RIPE NCC


All Rights Reserved

Permission to use, copy, modify, and distribute this software and its
documentation for any purpose and without fee is hereby granted, provided
that the above copyright notice appear in all copies and that both that
copyright notice and this permission notice appear in supporting
documentation, and that the name of the author not be used in advertising or
publicity pertaining to distribution of the software without specific,
written prior permission.

THE AUTHOR DISCLAIMS ALL WARRANTIES WITH REGARD TO THIS SOFTWARE, INCLUDING
ALL IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS; IN NO EVENT SHALL
AUTHOR BE LIABLE FOR ANY SPECIAL, INDIRECT OR CONSEQUENTIAL DAMAGES OR ANY
DAMAGES WHATSOEVER RESULTING FROM LOSS OF USE, DATA OR PROFITS, WHETHER IN
AN ACTION OF CONTRACT, NEGLIGENCE OR OTHER TORTIOUS ACTION, ARISING OUT OF
OR IN CONNECTION WITH THE USE OR PERFORMANCE OF THIS SOFTWARE. 

*/

/* 

Parts of this code have been engineered after analiyzing GNU Zebra's
source code and therefore might contain declarations/code from GNU
Zebra, Copyright (C) 1999 Kunihiro Ishiguro. Zebra is a free routing
software, distributed under the GNU General Public License. A copy of
this license is included with libbgpdump.

*/


/*
-------------------------------------------------------------------------------
Module Header
Filename          : bgdump_lib.c
Author            : Dan Ardelean (dan@ripe.net)
Date              : 02-SEP-2002
Revision          : 
Revised           : 
Description       : Library implementation
Language Version  : C
OSs Tested        : Linux 2.2.19
To Do             : 
-------------------------------------------------------------------------------
*/

#include "bgpdump_lib.h"
#include "bgpdump_mstream.h"

#include <sys/stat.h>
#include <fcntl.h>
#include <stdlib.h>
#include <unistd.h>

#include <netinet/in.h>
#include <sys/socket.h>
#include <arpa/inet.h>

    int process_mrtd_bgp(struct mstream *s,BGPDUMP_ENTRY *entry);
    int process_mrtd_table_dump(struct mstream *s,BGPDUMP_ENTRY *entry);
    int process_zebra_bgp(struct mstream *s,BGPDUMP_ENTRY *entry);
    int process_zebra_bgp_state_change(struct mstream *s,BGPDUMP_ENTRY *entry);
    
    int process_zebra_bgp_message(struct mstream *s,BGPDUMP_ENTRY *entry);
    int process_zebra_bgp_message_update(struct mstream *s,BGPDUMP_ENTRY *entry);

    int process_zebra_bgp_entry(struct mstream *s,BGPDUMP_ENTRY *entry);
    int process_zebra_bgp_snapshot(struct mstream *s,BGPDUMP_ENTRY *entry);

    void process_attr_init(BGPDUMP_ENTRY *entry);
    void process_attr_read(struct mstream *s,struct attr *attr);
    void process_attr_aspath_string(struct aspath *as);
    char aspath_delimiter_char (u_char type, u_char which);
    void process_attr_community_string(struct community *com);
    size_t strlcat(char *dst, const char *src, size_t size);

BGPDUMP *bgpdump_open_dump(char *filename) {
    BGPDUMP *this_dump=NULL;
    
    int fd;
    fd = open(filename, O_RDONLY);
    
    if(fd<0) return NULL;
    
    this_dump = malloc(sizeof(BGPDUMP));
    this_dump->fd=fd;
    strcpy(this_dump->filename, filename);
    this_dump->eof=0;

    return this_dump;
}

void bgpdump_close_dump(BGPDUMP *dump) {
    if(dump!=NULL) 
	close(dump->fd);
}

BGPDUMP_ENTRY*	bgpdump_read_next(BGPDUMP *dump) {
    BGPDUMP_ENTRY *this_entry=NULL;    
    struct mstream s;
    u_char	  *buffer;
    int		  ok=0;

    this_entry = malloc(sizeof(BGPDUMP_ENTRY));

    if(read(dump->fd,this_entry,12)!=12) { 
	free(this_entry);
	dump->eof=1; 
	return(NULL);
    }
    // Intel byte ordering stuff ...
    this_entry->type=ntohs(this_entry->type);
    this_entry->subtype=ntohs(this_entry->subtype);
    this_entry->time=ntohl(this_entry->time);
    this_entry->length=ntohl(this_entry->length);

    this_entry->attr=NULL;
        
    buffer = malloc(this_entry->length);
    if(read(dump->fd,buffer,this_entry->length)!=this_entry->length) { 
	free(this_entry); 
	free(buffer);
	dump->eof=1;
	return(NULL);
    }
        
    ok=0;
    mstream_init(&s,buffer,this_entry->length);

    switch(this_entry->type) {
	case BGPDUMP_TYPE_MRTD_BGP:		break;
	case BGPDUMP_TYPE_MRTD_TABLE_DUMP:	ok = process_mrtd_table_dump(&s,this_entry); break;
	case BGPDUMP_TYPE_ZEBRA_BGP:		ok = process_zebra_bgp(&s,this_entry); break;
    }

    free(buffer);
    if(!ok) {
	bgpdump_free_mem(this_entry);
	return NULL;
    }
    return this_entry;
}

void bgpdump_free_mem(BGPDUMP_ENTRY *entry) {
    if(entry!=NULL) {
	if(entry->attr != NULL) {
	    if(entry->attr->aspath != NULL) {
		if(entry->attr->aspath->data != NULL) free(entry->attr->aspath->data);
		if(entry->attr->aspath->str != NULL)  free(entry->attr->aspath->str);
		free(entry->attr->aspath);
	    }
	    
	    if(entry->attr->community != NULL) {
		if(entry->attr->community->val != NULL) free(entry->attr->community->val);
		if(entry->attr->community->str != NULL) free(entry->attr->community->str);
		free(entry->attr->community);
	    }

	    if(entry->attr->data != NULL) {
		free(entry->attr->data);
	    }

	    free(entry->attr);
	}


	switch(entry->type) {
	    case BGPDUMP_TYPE_MRTD_TABLE_DUMP:	
		break;
	    case BGPDUMP_TYPE_ZEBRA_BGP:
		switch(entry->subtype) {
		    case BGPDUMP_SUBTYPE_ZEBRA_BGP_MESSAGE:
			switch(entry->body.zebra_message.type) {
			    case BGP_MSG_UPDATE:
				if(entry->body.zebra_message.withdraw != NULL) free(entry->body.zebra_message.withdraw);
    				if(entry->body.zebra_message.announce != NULL) free(entry->body.zebra_message.announce);
				break;
			}
			break;
		}
		break;
	}

	free(entry);
    }
    
}

int process_mrtd_bgp(struct mstream *s,BGPDUMP_ENTRY *entry) {    
 return(1);
}

int process_mrtd_table_dump(struct mstream *s,BGPDUMP_ENTRY *entry) {

    mstream_getw(s,&entry->body.mrtd_table_dump.view);
    mstream_getw(s,&entry->body.mrtd_table_dump.sequence);
    mstream_get_ipv4(s,&entry->body.mrtd_table_dump.prefix.s_addr);
    mstream_getc(s,&entry->body.mrtd_table_dump.mask);
    mstream_getc(s,&entry->body.mrtd_table_dump.status);
    mstream_getl(s,(u_int32_t *)&entry->body.mrtd_table_dump.uptime);
    mstream_get_ipv4(s,&entry->body.mrtd_table_dump.peer_ip.s_addr);
    mstream_getw(s,&entry->body.mrtd_table_dump.peer_as);

    process_attr_init(entry);
    process_attr_read(s,entry->attr);
    
 return(1);
 
}

int process_zebra_bgp(struct mstream *s,BGPDUMP_ENTRY *entry) {
    switch(entry->subtype) {
	case BGPDUMP_SUBTYPE_ZEBRA_BGP_STATE_CHANGE:
	    return(process_zebra_bgp_state_change(s,entry));
	case BGPDUMP_SUBTYPE_ZEBRA_BGP_MESSAGE:
	    return(process_zebra_bgp_message(s,entry));
	case BGPDUMP_SUBTYPE_ZEBRA_BGP_ENTRY:
	    return(process_zebra_bgp_entry(s,entry));
	case BGPDUMP_SUBTYPE_ZEBRA_BGP_SNAPSHOT:
	    return(process_zebra_bgp_snapshot(s,entry));
	default:
	    return(0);    
    }
}


int process_zebra_bgp_state_change(struct mstream *s,BGPDUMP_ENTRY *entry) {

 return(0);
}

int process_zebra_bgp_message(struct mstream *s,BGPDUMP_ENTRY *entry) {

    mstream_getw(s,&entry->body.zebra_message.source_as);
    mstream_getw(s,&entry->body.zebra_message.destination_as);
    mstream_getw(s,&entry->body.zebra_message.interface_index);
    mstream_getw(s,&entry->body.zebra_message.address_family);
    mstream_get_ipv4(s,&entry->body.zebra_message.source_ip.s_addr);
    mstream_get_ipv4(s,&entry->body.zebra_message.destination_ip.s_addr);
    
    mstream_get (s, NULL, 16);		// read marker
    mstream_getw (s,&entry->body.zebra_message.size);
    mstream_getc (s,&entry->body.zebra_message.type);
    
    switch(entry->body.zebra_message.type) {    
	case BGP_MSG_OPEN:
	    // Not implemented yet
	    return(0);
	case BGP_MSG_UPDATE:
	    return(process_zebra_bgp_message_update(s,entry));
	case BGP_MSG_NOTIFY:
	    // Not implemented yet
	    return(0);
	case BGP_MSG_KEEPALIVE:
	    // Not implemented yet
	    return(0);
	case BGP_MSG_ROUTE_REFRESH_01:
	    // Not implemented yet
	    return(0);
	case BGP_MSG_ROUTE_REFRESH:
	    // Not implemented yet
	    return(0);
    }
    
    return(0);
}

int process_zebra_bgp_message_update(struct mstream *s,BGPDUMP_ENTRY *entry) {
    int withdraw_len;
    int announce_len;
    int countdown;
    int p_len;
    int p_bytes;

    withdraw_len = mstream_getw(s,NULL);
    countdown=withdraw_len;
    
    entry->body.zebra_message.withdraw_count=0;
    entry->body.zebra_message.withdraw=NULL;

    while(countdown > 0) {
	p_len = mstream_getc(s,NULL); countdown--;
	p_bytes = p_len / 8; if(p_len % 8 !=0) p_bytes++;
	
	entry->body.zebra_message.withdraw = realloc(entry->body.zebra_message.withdraw, 	
	    (entry->body.zebra_message.withdraw_count+1) * sizeof(struct prefix));

	memset(&entry->body.zebra_message.withdraw[entry->body.zebra_message.withdraw_count],0,sizeof(struct prefix));
	entry->body.zebra_message.withdraw[entry->body.zebra_message.withdraw_count].len=p_len;
	mstream_get(s,&entry->body.zebra_message.withdraw[entry->body.zebra_message.withdraw_count],p_bytes);
	countdown = countdown - p_bytes;

	entry->body.zebra_message.withdraw_count++;
    }
    
    process_attr_init(entry);
    process_attr_read(s,entry->attr);

    entry->body.zebra_message.announce_count=0;
    entry->body.zebra_message.announce=NULL;

    announce_len = entry->body.zebra_message.size - 23 - withdraw_len - entry->attr->len;
    countdown = announce_len;
    
    while(countdown > 0) {
	p_len = mstream_getc(s,NULL); countdown--;
	p_bytes = p_len / 8; if(p_len % 8 !=0) p_bytes++;

	entry->body.zebra_message.announce = realloc(entry->body.zebra_message.announce,
	    (entry->body.zebra_message.announce_count+1) * sizeof(struct prefix));

	memset(&entry->body.zebra_message.announce[entry->body.zebra_message.announce_count],0,sizeof(struct prefix));
	entry->body.zebra_message.announce[entry->body.zebra_message.announce_count].len=p_len;
	mstream_get(s,&entry->body.zebra_message.announce[entry->body.zebra_message.announce_count],p_bytes);
	countdown = countdown - p_bytes;

	entry->body.zebra_message.announce_count++;
    }

    return(1);
}

int process_zebra_bgp_entry(struct mstream *s,BGPDUMP_ENTRY *entry) {

 return(0);
}

int process_zebra_bgp_snapshot(struct mstream *s,BGPDUMP_ENTRY *entry) {

 return(0);
}

void process_attr_init(BGPDUMP_ENTRY *entry) {
    entry->attr = malloc(sizeof(struct attr));
    
    entry->attr->refcnt		= 0;
    entry->attr->flag		= 0;
    entry->attr->origin		= -1;
    entry->attr->nexthop.s_addr	= -1;
    entry->attr->med		= -1;
    entry->attr->local_pref	= -1;
    entry->attr->aggregator_as 	= -1;
    entry->attr->aggregator_addr.s_addr = -1;
    entry->attr->weight		= -1;
    entry->attr->cluster	= NULL;
    
    entry->attr->mp_nexthop_len			= 0;
    entry->attr->mp_nexthop_global_in.s_addr	= -1;
    entry->attr->mp_nexthop_local_in.s_addr	= -1;
    
    entry->attr->aspath		= NULL;
    entry->attr->community	= NULL;
    entry->attr->transit	= NULL;
    entry->attr->len		= 0;
    entry->attr->data		= NULL;
}

void process_attr_read(struct mstream *s,struct attr *attr) {
    u_char	flag;
    u_char	type;
    u_int32_t 	len;
    u_int32_t	end;
    
    mstream_getw(s,&attr->len);
    end = s->position+attr->len;

    attr->data=malloc(attr->len);
    memcpy(attr->data,&s->start[s->position],attr->len);
    
    while(s->position < end) {
	mstream_getc(s,&flag);
	mstream_getc(s,&type);
	
	if(flag & BGP_ATTR_FLAG_EXTLEN) 
	    len=mstream_getw(s,NULL);
	else 
	    len=mstream_getc(s,NULL);
	
	switch(type) {
	    case BGP_ATTR_ORIGIN:
		attr->flag = attr->flag | ATTR_FLAG_BIT (BGP_ATTR_ORIGIN);
		mstream_getc(s,&attr->origin);
		break;	    
	    case BGP_ATTR_AS_PATH:
		attr->flag = attr->flag | ATTR_FLAG_BIT (BGP_ATTR_AS_PATH);
		attr->aspath		= malloc(sizeof(struct aspath));
		attr->aspath->refcnt	= 0;
		attr->aspath->length	= len;
		attr->aspath->count	= 0;
		attr->aspath->data	= malloc(len);
		mstream_get(s,attr->aspath->data,len);
		attr->aspath->str	= NULL;
		process_attr_aspath_string(attr->aspath);
		break;
	    case BGP_ATTR_NEXT_HOP:
		attr->flag = attr->flag | ATTR_FLAG_BIT (BGP_ATTR_NEXT_HOP);
		mstream_get_ipv4(s,&attr->nexthop.s_addr);
		break;
	    case BGP_ATTR_MULTI_EXIT_DISC:
		attr->flag = attr->flag | ATTR_FLAG_BIT (BGP_ATTR_MULTI_EXIT_DISC);
		mstream_getl(s,&attr->med);
		break;
	    case BGP_ATTR_LOCAL_PREF:
		attr->flag = attr->flag | ATTR_FLAG_BIT (BGP_ATTR_LOCAL_PREF);
		mstream_getl(s,&attr->local_pref);
		break;
	    case BGP_ATTR_ATOMIC_AGGREGATE:
		attr->flag = attr->flag | ATTR_FLAG_BIT (BGP_ATTR_ATOMIC_AGGREGATE);
		break;
	    case BGP_ATTR_AGGREGATOR:
		attr->flag = attr->flag | ATTR_FLAG_BIT (BGP_ATTR_AGGREGATOR);
		mstream_getw(s,&attr->aggregator_as);
		mstream_get_ipv4(s,&attr->aggregator_addr.s_addr);
		break;
	    case BGP_ATTR_COMMUNITIES:
		attr->flag = attr->flag | ATTR_FLAG_BIT (BGP_ATTR_COMMUNITIES);
		attr->community		= malloc(sizeof(struct community));
		attr->community->refcnt = 0;
		attr->community->size	= len / 4;
		attr->community->val	= malloc(len);
		mstream_get(s,attr->community->val,len);
		attr->community->str	= NULL;
		process_attr_community_string(attr->community);
		break;
	    default:
		mstream_get(s,NULL,len);
	}
    }
}

void process_attr_aspath_string(struct aspath *as) {
  int space;
  u_char type;
  caddr_t pnt;
  caddr_t end;
  struct assegment *assegment;
  int str_size = ASPATH_STR_DEFAULT_LEN;
  int str_pnt;
  u_char *str_buf;
  int count = 0;

  /* Empty aspath. */
  if (as->length == 0)
    {
      str_buf = malloc(1);
      str_buf[0] = '\0';
      as->count = count;
      as->str = str_buf;
      return;
    }

  /* Set default value. */
  space = 0;
  type = AS_SEQUENCE;

  /* Set initial pointer. */
  pnt = as->data;
  end = pnt + as->length;

  str_buf = malloc(str_size);
  str_pnt = 0;

  assegment = (struct assegment *) pnt;

  while (pnt < end)
    {
      int i;
      int estimate_len;

      /* For fetch value. */
      assegment = (struct assegment *) pnt;

      /* Check AS type validity. */
      if ((assegment->type != AS_SET) && 
	  (assegment->type != AS_SEQUENCE) &&
	  (assegment->type != AS_CONFED_SET) && 
	  (assegment->type != AS_CONFED_SEQUENCE))
	{
	  free(str_buf);
	  str_buf=malloc(strlen(ASPATH_STR_ERROR)+1);
	  strcpy(str_buf,ASPATH_STR_ERROR);
	  as->count=0;
	  as->str = str_buf;
	  return;
	}

      /* Check AS length. */
      if ((pnt + (assegment->length * AS_VALUE_SIZE) + AS_HEADER_SIZE) > end)
	{
	  free(str_buf);
	  str_buf=malloc(strlen(ASPATH_STR_ERROR)+1);
	  strcpy(str_buf,ASPATH_STR_ERROR);
	  as->count=0;
	  as->str = str_buf;
	  return;
	}

      /* Buffer length check. */
      estimate_len = ((assegment->length * 6) + 4);
      
      /* String length check. */
      while (str_pnt + estimate_len >= str_size)
	{
	  str_size *= 2;
	  str_buf = realloc (str_buf, str_size);
	}

      /* If assegment type is changed, print previous type's end
         character. */
      if (type != AS_SEQUENCE)
	str_buf[str_pnt++] = aspath_delimiter_char (type, AS_SEG_END);
      if (space)
	str_buf[str_pnt++] = ' ';

      if (assegment->type != AS_SEQUENCE)
	str_buf[str_pnt++] = aspath_delimiter_char (assegment->type, AS_SEG_START);

      space = 0;

      /* Increment count - ignoring CONFED SETS/SEQUENCES */
      if (assegment->type != AS_CONFED_SEQUENCE
	  && assegment->type != AS_CONFED_SET)
	{
	  if (assegment->type == AS_SEQUENCE)
	    count += assegment->length;
	  else if (assegment->type == AS_SET)
	    count++;
	}

      for (i = 0; i < assegment->length; i++)
	{
	  int len;

	  if (space)
	    {
	      if (assegment->type == AS_SET
		  || assegment->type == AS_CONFED_SET)
		str_buf[str_pnt++] = ',';
	      else
		str_buf[str_pnt++] = ' ';
	    }
	  else
	    space = 1;

	  len = sprintf (str_buf + str_pnt, "%d", ntohs (assegment->asval[i]));
	  str_pnt += len;
	}

      type = assegment->type;
      pnt += (assegment->length * AS_VALUE_SIZE) + AS_HEADER_SIZE;
    }

  if (assegment->type != AS_SEQUENCE)
    str_buf[str_pnt++] = aspath_delimiter_char (assegment->type, AS_SEG_END);

  str_buf[str_pnt] = '\0';

  as->count = count;
  as->str = malloc(strlen(str_buf)+1);
  strcpy(as->str,str_buf);
  free(str_buf);
}

char aspath_delimiter_char (u_char type, u_char which) {
  int i;
  struct
  {
    int type;
    char start;
    char end;
  } aspath_delim_char [] =
    {
      { AS_SET,             '{', '}' },
      { AS_SEQUENCE,        ' ', ' ' },
      { AS_CONFED_SET,      '[', ']' },
      { AS_CONFED_SEQUENCE, '(', ')' },
      { 0 }
    };

  for (i = 0; aspath_delim_char[i].type != 0; i++)
    {
      if (aspath_delim_char[i].type == type)
	{
	  if (which == AS_SEG_START)
	    return aspath_delim_char[i].start;
	  else if (which == AS_SEG_END)
	    return aspath_delim_char[i].end;
	}
    }
  return ' ';
}


void process_attr_community_string(struct community *com) {

  char buf[BUFSIZ];
  int i;
  u_int32_t comval;
  u_int16_t as;
  u_int16_t val;

  memset (buf, 0, BUFSIZ);

  for (i = 0; i < com->size; i++) 
    {
      memcpy (&comval, com_nthval (com, i), sizeof (u_int32_t));
      comval = ntohl (comval);
      switch (comval) 
	{
	case COMMUNITY_NO_EXPORT:
	  strlcat (buf, " no-export", BUFSIZ);
	  break;
	case COMMUNITY_NO_ADVERTISE:
	  strlcat (buf, " no-advertise", BUFSIZ);
	  break;
	case COMMUNITY_LOCAL_AS:
	  strlcat (buf, " local-AS", BUFSIZ);
	  break;
	default:
	  as = (comval >> 16) & 0xFFFF;
	  val = comval & 0xFFFF;
	  snprintf (buf + strlen (buf), BUFSIZ - strlen (buf), 
		    " %d:%d", as, val);
	  break;
	}
    }

    com->str = malloc(strlen(buf)+1);
    strcpy(com->str, buf);
}

size_t strlcat(char *dst, const char *src, size_t size) {
  if (strlen (dst) + strlen (src) >= size)
    return -1;

  strcat (dst, src);

  return (strlen(dst));
}
